@@ -1,5 +1,5 @@
 /*
- * Copyright 1998-2002 by Albert Cahalan; all rights reserved.
+ * Copyright 1998-2003 by Albert Cahalan; all rights reserved.
  * This file may be used subject to the terms and conditions of the
  * GNU Library General Public License Version 2, or any later version
  * at your option, as published by the Free Software Foundation.
@@ -1130,8 +1130,17 @@ try_bsd:
   err2 = select_bits_setup();
   if(err2) goto total_failure;
 
+  // Feel a need to patch this out? First of all, read the FAQ.
+  // Second of all, talk to me. Without this warning, people can
+  // get seriously confused. Ask yourself if users would freak out
+  // about "ps -aux" suddenly changing behavior if a user "x" were
+  // added to the system.
   if(!(personality & PER_FORCE_BSD))
-    fprintf(stderr, "Bad syntax, perhaps a bogus '-'?\n");
+    fprintf(stderr, "Warning: bad syntax, perhaps a bogus '-'? See http://procps.sf.net/faq.html\n");
+  // Remember: contact albert@users.sf.net or procps-feedback@lists.sf.net
+  // if you should feel tempted. Be damn sure you understand all
+  // the issues. The same goes for other stuff too, BTW. Please ask.
+  // I'm happy to justify various implementation choices.
 
   choose_dimensions();
   return 0;
